"""
A requests-based HTTP client for interacting with containers that have
forwarded ports.
"""

import hyperlink

import requests


def _path_segments(url, path_str):
    # Absolute path
    if path_str.startswith('/'):
        return path_str[1:].split('/')
    # Relative path
    return url.child(*path_str.split('/')).path


class ContainerHttpClient:
    """
    HTTP client for a specific container.

    In most cases, these should be obtained from
    :meth:`.ContainerDefinition.http_client` instead of being instantiated
    directly.
    """

    URL_DEFAULTS = {'scheme': 'http'}

    def __init__(self, host, port, url_defaults=None, session=None):
        """
        :param host:
            The address for the host to connect to.
        :param port:
            The port for the host to connect to.
        :param dict url_defaults:
            Parameters to default to in the generated URLs, see
            `~hyperlink.URL`.
        :param session:
            A Requests' Session object (or something like it).
        """
        if session is None:
            session = requests.Session()
        self._session = session

        _url_defaults = self.URL_DEFAULTS.copy()
        if url_defaults is not None:
            _url_defaults.update(url_defaults)
        self._base_url = hyperlink.URL(
            host=host, port=int(port), **_url_defaults)

    def __enter__(self):
        return self

    def close(self):
        """
        Closes the underlying Session object.
        """
        self._session.close()

    def __exit__(self, *args):
        self.close()

    @classmethod
    def for_container(cls, container, container_port=None):
        """
        :param container:
            The container to make requests against.
        :param container_port:
            The container port to make requests against. If ``None``, the first
            container port is used.
        :returns:
            A ContainerClient object configured to make requests to the
            container.
        """
        if container_port is not None:
            host, port = container.get_host_port(container_port)
        else:
            host, port = container.get_first_host_port()

        return cls(host, port)

    def _url(self, path, kwargs):
        kwargs = kwargs if kwargs is not None else {}
        if path is not None:
            kwargs['path'] = _path_segments(self._base_url, path)
        return self._base_url.replace(**kwargs).to_text()

    def request(self, method, path=None, url_kwargs=None, **kwargs):
        """
        Make a request against a container.

        :param method:
            The HTTP method to use.
        :param list path:
            The HTTP path (either absolute or relative).
        :param dict url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param kwargs:
            Any other parameters to pass to Requests.
        """
        return self._session.request(
            method, self._url(path, url_kwargs), **kwargs)

    def get(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a GET request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param \*\*kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.get(self._url(path, url_kwargs), **kwargs)

    def options(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends an OPTIONS request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param \*\*kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.options(self._url(path, url_kwargs), **kwargs)

    def head(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a HEAD request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param \*\*kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.head(self._url(path, url_kwargs), **kwargs)

    def post(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a POST request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param \*\*kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.post(self._url(path, url_kwargs), **kwargs)

    def put(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a PUT request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param \*\*kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.put(self._url(path, url_kwargs), **kwargs)

    def patch(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a PUT request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param \*\*kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.patch(self._url(path, url_kwargs), **kwargs)

    def delete(self, path=None, url_kwargs=None, **kwargs):
        """
        Sends a PUT request.

        :param path:
            The HTTP path (either absolute or relative).
        :param url_kwargs:
            Parameters to override in the generated URL. See `~hyperlink.URL`.
        :param \*\*kwargs:
            Optional arguments that ``request`` takes.
        :return: response object
        """
        return self._session.delete(self._url(path, url_kwargs), **kwargs)



"""
Tools for waiting on and matching log lines from a container.
"""

import re
from abc import ABC, abstractmethod

from ._lowlevel import stream_logs


def output_lines(raw_output, encoding='utf-8'):
    """
    Convert bytestring container output into a sequence of unicode lines.

    :param bytestring raw_output: Container output bytes.
    :param encoding: The encoding to use when converting bytes to unicode
        (default ``utf-8``).

    :returns: list[str]
    """
    return raw_output.decode(encoding).splitlines()


def _last_few_log_lines(container):
    return container.logs(tail=100).decode('utf-8')


def wait_for_logs_matching(container, matcher, timeout=10, encoding='utf-8',
                           **logs_kwargs):
    """
    Wait for matching log line(s) from the given container by streaming the
    container's stdout and/or stderr outputs.

    Each log line is decoded and any trailing whitespace is stripped before the
    line is matched.

    :param ~docker.models.containers.Container container:
        Container who's log lines to wait for.
    :param matcher:
        Callable that returns True once it has matched a decoded log line(s).
    :param timeout:
        Timeout value in seconds.
    :param encoding:
        Encoding to use when decoding container output to strings.
    :param logs_kwargs:
        Additional keyword arguments to pass to ``container.logs()``. For
        example, the ``stdout`` and ``stderr`` boolean arguments can be used to
        determine whether to stream stdout or stderr or both (the default).

    :returns:
        The final matching log line.
    :raises TimeoutError:
        When the timeout value is reached before matching log lines have been
        found.
    :raises RuntimeError:
        When all log lines have been consumed but matching log lines have not
        been found (the container must have stopped for its stream to have
        ended without error).
    """
    try:
        for line in stream_logs(
                container, timeout=timeout, **logs_kwargs):
            # Drop the trailing newline
            line = line.decode(encoding).rstrip()
            if matcher(line):
                return line
    except TimeoutError:
        raise TimeoutError('\n'.join([
            ('Timeout ({}s) waiting for logs matching {}.'.format(
                timeout, matcher)),
            'Last few log lines:',
            _last_few_log_lines(container),
        ]))

    raise RuntimeError('\n'.join([
        'Logs matching {} not found.'.format(matcher),
        'Last few log lines:',
        _last_few_log_lines(container),
    ]))


class LogMatcher(ABC):
    """
    Abstract base class for log matchers.
    """

    @abstractmethod
    def match(self, log_line):
        """
        Return ``True`` if the matcher matches a line, otherwise ``False``.
        """

    @abstractmethod
    def args_str(self):
        """
        Return an args string for the repr.
        """

    def __call__(self, log_line):
        return self.match(log_line)

    def __str__(self):
        return '{}({})'.format(type(self).__name__, self.args_str())

    def __repr__(self):
        return str(self)


def to_matcher(matcher_factory, obj):
    """
    Turn an object into a :class:`LogMatcher` unless it already is one.

    :param matcher_factory: A callable capable of turning `obj` into a
        :class:`LogMatcher`.
    :param obj: A :class:`LogMatcher` or an object to turn into one.

    :returns: :class:`LogMatcher`
    """
    return obj if isinstance(obj, LogMatcher) else matcher_factory(obj)


class CombinationLogMatcher(LogMatcher):
    """
    Matcher that combines multiple input matchers.
    """
    def __init__(self, *matchers):
        self._matchers = matchers

    @classmethod
    def by_equality(cls, *expected_lines):
        """
        Construct an instance of this combination matcher from a list of
        expected log lines and/or LogMatcher instances.
        """
        return cls(*(to_matcher(EqualsMatcher, l) for l in expected_lines))

    @classmethod
    def by_regex(cls, *patterns):
        """
        Construct an instance of this combination matcher from a list of
        regex patterns and/or LogMatcher instances.
        """
        return cls(*(to_matcher(RegexMatcher, p) for p in patterns))


class OrderedLinesMatcher(CombinationLogMatcher):
    """
    Matcher that takes a list of matchers, and uses one after the next after
    each has a successful match. Returns True ("matches") on the final match.

    **Note:** This is a *stateful* matcher. Once it has done its matching,
    you'll need to create a new instance.
    """
    def __init__(self, *matchers):
        super().__init__(*matchers)
        self._position = 0

    def match(self, log_line):
        """
        Return ``True`` if the expected matchers are matched in the expected
        order, otherwise ``False``.
        """
        if self._position == len(self._matchers):
            raise RuntimeError('Matcher exhausted, no more matchers to use')

        matcher = self._matchers[self._position]
        if matcher(log_line):
            self._position += 1

        if self._position == len(self._matchers):
            # All patterns have been matched
            return True

        return False

    def args_str(self):
        """
        Return an args string for the repr.
        """
        matched = [str(m) for m in self._matchers[:self._position]]
        unmatched = [str(m) for m in self._matchers[self._position:]]
        return 'matched=[{}], unmatched=[{}]'.format(
            ', '.join(matched), ', '.join(unmatched))


class UnorderedLinesMatcher(CombinationLogMatcher):
    """
    Matcher that takes a list of matchers, and matches each one to a line. Each
    line is tested against each unmatched matcher until a match is found or all
    unmatched matchers are checked. Returns True ("matches") on the final
    match.

    .. note::

        This is a *stateful* matcher. Once it has done its matching,
        you'll need to create a new instance.
    """
    def __init__(self, *matchers):
        super().__init__(*matchers)
        self._used_matchers = []

    @property
    def _unused_matchers(self):
        return [m for m in self._matchers if m not in self._used_matchers]

    def match(self, log_line):
        """
        Return ``True`` if the expected matchers are matched in any order,
        otherwise ``False``.
        """
        if not self._unused_matchers:
            raise RuntimeError('Matcher exhausted, no more matchers to use')

        for matcher in self._unused_matchers:
            if matcher(log_line):
                self._used_matchers.append(matcher)
                break

        if not self._unused_matchers:
            # All patterns have been matched
            return True

        return False

    def args_str(self):
        """
        Return an args string for the repr.
        """
        matched = [str(m) for m in self._used_matchers]
        unmatched = [str(m) for m in self._unused_matchers]
        return 'matched=[{}], unmatched=[{}]'.format(
            ', '.join(matched), ', '.join(unmatched))


class EqualsMatcher(LogMatcher):
    """
    Matcher that matches log lines by equality.
    """
    def __init__(self, expected_line):
        self._expected_line = expected_line

    def match(self, log_line):
        """
        Return ``True`` if the log line matches the expected value exactly,
        otherwise ``False``.
        """
        return log_line == self._expected_line

    def args_str(self):
        """
        Return an args string for the repr.
        """
        return repr(self._expected_line)


class RegexMatcher(LogMatcher):
    """
    Matcher that matches log lines by regex pattern.
    """
    def __init__(self, pattern):
        self._regex = re.compile(pattern)

    def match(self, log_line):
        """
        Return ``True`` if the log line matches the expected regex, otherwise
        ``False``.
        """
        return self._regex.search(log_line) is not None

    def args_str(self):
        """
        Return an args string for the repr.
        """
        return repr(self._regex.pattern)


# Members of this module are documented in the order they appear here.
__all__ = [
    # Matchers
    'LogMatcher',

    'CombinationLogMatcher',
    'EqualsMatcher',
    'RegexMatcher',
    'OrderedLinesMatcher',
    'UnorderedLinesMatcher',

    # Functions
    'output_lines',
    'to_matcher',
    'wait_for_logs_matching',
]



"""
A number of pytest fixtures or factories for fixtures.
"""

import os

import pytest

from seaworthy.helpers import DockerHelper


def docker_helper_fixture(name='docker_helper', scope='module', **kwargs):
    """
    Create a fixture for :class:`~seaworthy.DockerHelper`.

    This can be used to create a fixture with a different name to the default.
    It can also be used to override the scope of the default fixture::

        docker_helper = docker_helper_fixture(scope='class')

    :param name: The name of the fixture.
    :param scope: The scope of the fixture.
    :param kwargs:
        Keyword arguments to pass to the :class:`~seaworthy.DockerHelper`
        constructor.
    """
    @pytest.fixture(name=name, scope=scope)
    def fixture():
        namespace = kwargs.pop('namespace', 'test')
        if 'PYTEST_XDIST_WORKER' in os.environ:  # pragma: no cover
            namespace = '{}_{}'.format(
                namespace, os.environ['PYTEST_XDIST_WORKER'])
        docker_helper = DockerHelper(namespace=namespace, **kwargs)
        yield docker_helper
        docker_helper.teardown()
    return fixture


#: Default fixture for :class:`~seaworthy.DockerHelper`. Has module scope.
docker_helper = docker_helper_fixture()


def image_fetch_fixture(image, name, scope='module'):
    """
    Create a fixture to fetch an image.
    """
    @pytest.fixture(name=name, scope=scope)
    def fixture(docker_helper):
        return docker_helper.images.fetch(image)
    return fixture


def resource_fixture(definition, name, scope='function'):
    """
    Create a fixture for a resource.

    .. note:: This function returns a fixture function. It is important to keep
        a reference to the returned function within the scope of the tests that
        use the fixture.

    .. code-block:: python

        fixture = resource_fixture(PostgreSQLContainer(), 'postgresql')

        def test_container(postgresql):
            \"""Test something about the PostgreSQL container...\"""

    :param definition:
        A resource definition, one of those defined in the
        :mod:`seaworthy.definitions` module.
    :param name: The fixture name.
    :param scope: The scope of the fixture.

    :returns: The fixture function.
    """
    @pytest.fixture(name=name, scope=scope)
    def fixture(docker_helper):
        definition.setup(helper=docker_helper)
        yield definition
        definition.teardown()

    return fixture


def _clean_container_fixture(name, raw_name):
    @pytest.fixture(name=name)
    def clean_fixture(request):
        container = request.getfixturevalue(raw_name)
        if 'clean_{}'.format(name) in request.keywords:
            container.clean()
        return container

    return clean_fixture


def clean_container_fixtures(container, name, scope='class'):
    """
    Creates a fixture for a container that can be "cleaned". When a code block
    is marked with ``@pytest.mark.clean_<fixture name>`` then the ``clean``
    method will be called on the container object before it is passed as an
    argument to the test function.

    .. note:: This function returns two fixture functions. It is important to
        keep references to the returned functions within the scope of the tests
        that use the fixtures.

    .. code-block:: python

        f1, f2 = clean_container_fixtures(PostgreSQLContainer(), 'postgresql')

        class TestPostgresqlContainer
            @pytest.mark.clean_postgresql
            def test_clean_container(self, web_container, postgresql):
                \"""
                Test something about the container that requires it to have a
                clean state (e.g. database table creation).
                \"""

            def test_dirty_container(self, web_container, postgresql):
                \"""
                Test something about the container that doesn't require it to
                have a clean state (e.g. testing something about a dependent
                container).
                \"""

    :param container:
        A "container" object that is a subclass of
        :class:`.ContainerDefinition`.
    :param name:
        The fixture name.
    :param scope:
        The scope of the fixture.

    :returns:
        A tuple of two fixture functions.
    """
    raw_name = 'raw_{}'.format(name)
    return (resource_fixture(container, raw_name, scope),
            _clean_container_fixture(name, raw_name))


__all__ = ['clean_container_fixtures', 'docker_helper',
           'docker_helper_fixture', 'image_fetch_fixture', 'resource_fixture']



import socket
import struct
import threading
import unittest
from collections import namedtuple
from datetime import datetime

from seaworthy.checks import docker_client, dockertest
from seaworthy.helpers import DockerHelper, fetch_images
from seaworthy.logs import (
    EqualsMatcher, OrderedLinesMatcher, RegexMatcher, UnorderedLinesMatcher,
    stream_logs, wait_for_logs_matching)

# We use this image to test with because it is a small (~4MB) image from
# https://github.com/docker-library/official-images that we can run shell
# scripts in.
IMG = 'alpine:latest'


@dockertest()
def setUpModule():  # noqa: N802 (The camelCase is mandated by unittest.)
    with docker_client() as client:
        fetch_images(client, [IMG])


class TestEqualsMatcher(unittest.TestCase):
    def test_matching(self):
        """ Matches exactly equal strings and nothing else. """
        matcher = EqualsMatcher('foo')
        self.assertTrue(matcher('foo'))
        self.assertFalse(matcher('foobar'))

    def test_str(self):
        """ The string representation is readable. """
        matcher = EqualsMatcher('bar')
        self.assertEqual(str(matcher), "EqualsMatcher('bar')")
        self.assertEqual(repr(matcher), str(matcher))


class TestRegexMatcher(unittest.TestCase):
    def test_matching(self):
        """
        Matches strings that match the pattern, doesn't match other strings.
        """
        matcher = RegexMatcher(r'^foo')
        self.assertTrue(matcher('foobar'))
        self.assertFalse(matcher('barfoo'))

    def test_str(self):
        """ The string representation is readable. """
        matcher = RegexMatcher(r'^bar')
        self.assertEqual(str(matcher), "RegexMatcher('^bar')")
        self.assertEqual(repr(matcher), str(matcher))


class TestOrderedLinesMatcher(unittest.TestCase):
    def test_matching(self):
        """
        Applies each matcher sequentially and returns True on the final match.
        """
        matcher = OrderedLinesMatcher(
            EqualsMatcher('foo'), RegexMatcher('^bar'))

        self.assertFalse(matcher('barfoo'))
        self.assertFalse(matcher('foo'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('barfoo'))

    def test_by_equality(self):
        """
        The ``by_equality`` utility method takes a list of strings and produces
        a matcher that matches those strings by equality sequentially.
        """
        matcher = OrderedLinesMatcher.by_equality('foo', 'bar')

        self.assertFalse(matcher('bar'))
        self.assertFalse(matcher('foo'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('bar'))

    def test_by_regex(self):
        """
        The ``by_regex`` utility method takes a list of patterns and produces a
        matcher that matches by those regex patterns sequentially.
        """
        matcher = OrderedLinesMatcher.by_regex(r'^foo', r'bar$')

        self.assertFalse(matcher('fuzzbar'))
        self.assertFalse(matcher('foobar'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('foobar'))

    def test_exhaustion(self):
        """
        Once all matchers have been matched, further calls to ``match`` should
        raise an error.
        """
        matcher = OrderedLinesMatcher.by_equality('foo')
        self.assertTrue(matcher('foo'))

        with self.assertRaises(RuntimeError) as cm:
            matcher('bar')
        self.assertEqual(str(cm.exception),
                         'Matcher exhausted, no more matchers to use')

    def test_str(self):
        """
        The string representation is readable and shows which matchers have
        been matched and which are still to be matched.
        """
        matcher = OrderedLinesMatcher(
            EqualsMatcher('foo'), RegexMatcher('^bar'))

        self.assertEqual(
            str(matcher),
            'OrderedLinesMatcher(matched=[], '
            "unmatched=[EqualsMatcher('foo'), RegexMatcher('^bar')])")
        self.assertEqual(repr(matcher), str(matcher))

        self.assertFalse(matcher('foo'))

        self.assertEqual(
            str(matcher),
            "OrderedLinesMatcher(matched=[EqualsMatcher('foo')], "
            "unmatched=[RegexMatcher('^bar')])")
        self.assertEqual(repr(matcher), str(matcher))

        self.assertTrue(matcher('barfoo'))

        self.assertEqual(
            str(matcher),
            "OrderedLinesMatcher(matched=[EqualsMatcher('foo'), "
            "RegexMatcher('^bar')], unmatched=[])")
        self.assertEqual(repr(matcher), str(matcher))


class TestUnorderedLinesMatcher(unittest.TestCase):
    def test_matching(self):
        """
        Applies each matcher sequentially and returns True on the final match.
        """
        matcher = UnorderedLinesMatcher(
            EqualsMatcher('foo'), RegexMatcher('^bar'))

        self.assertFalse(matcher('barfoo'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('foo'))

        matcher = UnorderedLinesMatcher(
            EqualsMatcher('foo'), RegexMatcher('^bar'))

        self.assertFalse(matcher('foo'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('barfoo'))

    def test_by_equality(self):
        """
        The ``by_equality`` utility method takes a list of strings and produces
        a matcher that matches those strings by equality sequentially.
        """
        matcher = UnorderedLinesMatcher.by_equality('foo', 'bar')

        self.assertFalse(matcher('foo'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('bar'))

    def test_by_regex(self):
        """
        The ``by_regex`` utility method takes a list of patterns and produces a
        matcher that matches by those regex patterns sequentially.
        """
        matcher = UnorderedLinesMatcher.by_regex(r'^foo', r'bar$')

        self.assertFalse(matcher('foobar'))
        self.assertFalse(matcher('baz'))
        self.assertTrue(matcher('foobar'))

    def test_exhaustion(self):
        """
        Once all matchers have been matched, further calls to ``match`` should
        raise an error.
        """
        matcher = UnorderedLinesMatcher.by_equality('foo')
        self.assertTrue(matcher('foo'))

        with self.assertRaises(RuntimeError) as cm:
            matcher('bar')
        self.assertEqual(str(cm.exception),
                         'Matcher exhausted, no more matchers to use')

    def test_str(self):
        """
        The string representation is readable and shows which matchers have
        been matched and which are still to be matched.
        """
        matcher = UnorderedLinesMatcher(
            EqualsMatcher('foo'), RegexMatcher('^bar'))

        self.assertEqual(
            str(matcher),
            'UnorderedLinesMatcher(matched=[], '
            "unmatched=[EqualsMatcher('foo'), RegexMatcher('^bar')])")
        self.assertEqual(repr(matcher), str(matcher))

        self.assertFalse(matcher('foo'))

        self.assertEqual(
            str(matcher),
            "UnorderedLinesMatcher(matched=[EqualsMatcher('foo')], "
            "unmatched=[RegexMatcher('^bar')])")
        self.assertEqual(repr(matcher), str(matcher))

        self.assertTrue(matcher('barfoo'))

        self.assertEqual(
            str(matcher),
            "UnorderedLinesMatcher(matched=[EqualsMatcher('foo'), "
            "RegexMatcher('^bar')], unmatched=[])")
        self.assertEqual(repr(matcher), str(matcher))


class FakeLogsContainer:
    """
    A container object stub that emits canned logs.

    Logs can either be streamed through a socket using ``.attach_socket()`` or
    tailed by using ``.logs()``. After a log entry has been streamed, it is
    stored and we don't stream it next time. Only logs that have already been
    streamed are tailed.
    """

    def __init__(self, log_entries, expected_params=None, close_timeout=2):
        self.log_entries = log_entries
        self._seen_logs = []
        self._expected_stream_params = {
            'stdout': True,
            'stderr': True,
            'follow': True,
        }
        if expected_params is not None:
            self._expected_stream_params.update(expected_params)
        self._feeder = None
        self._client_sockets = set()
        self._api = FakeAPIClient()
        self._close_timeout = close_timeout
        self.client = namedtuple('DockerClient', 'api')(self._api)

    def cleanup(self):
        self.cancel_feeder()
        while self._client_sockets:
            self._client_sockets.pop().close()
        assert self._api._multiplexed_response_stream_helper == 'ORIG_MRSH'

    def cancel_feeder(self):
        feeder = self._feeder
        if feeder is not None:
            feeder.finished.set()
            feeder.join()

    def logs(self, stream=False, **kw):
        tail = kw.pop('tail', 'all')
        if stream:
            return self._stream_logs(tail, kw)
        else:
            return b''.join(self._tail_logs(tail))

    def _tail_logs(self, tail):
        if tail == 0:
            # Nothing to tail.
            return []
        if tail == 'all':
            # ALL THE LOGS!
            return self._seen_logs
        # Just some of the logs.
        assert tail > 0
        return self._seen_logs[-tail:]

    def _stream_logs(self, tail, kw):
        # Check that we're properly monkeypatched.
        assert self._api._multiplexed_response_stream_helper != 'ORIG_MRSH'
        assert self._feeder is None
        assert kw == self._expected_stream_params
        server, client = socket.socketpair()
        self._client_sockets.add(client)
        self._feeder = LogFeeder(self, server, self._tail_logs(tail))
        # Add a raw attr for the client to read from.
        self._feeder.raw = socket.SocketIO(client, 'rb')
        self._feeder.start()
        return self._feeder, client


class FakeAPIClient:
    """
    A fake APIClient so we can make sure our monkeypatch is properly applied
    and removed.
    """
    def __init__(self):
        self._multiplexed_response_stream_helper = 'ORIG_MRSH'


class LogFeeder(threading.Thread):
    def __init__(self, container, sock, tail):
        super().__init__()
        self.con = container
        self.sock = sock
        self.tail = tail
        self.finished = threading.Event()

    def close(self):
        # This is a bit of a hack to avoid having a separate fake request
        # object for the streaming client to close when it's done.
        self.con.cleanup()

    def send_line(self, line):
        data = b'\x00\x00\x00\x00' + struct.pack('>L', len(line)) + line
        self.sock.send(data)

    def run(self):
        # Emit tailed lines.
        for line in self.tail:
            self.send_line(line)
        # Emit previously unstreamed lines at designated intervals.
        for delay, line in self.con.log_entries[len(self.con._seen_logs):]:
            # Wait for either cancelation (break) or timeout (no break).
            if self.finished.wait(delay):
                break
            self.con._seen_logs.append(line)
            self.send_line(line)
        # Wait until we're done, which we may already be. Since some of the
        # tests don't do client-side timeouts, we use a fake-specific "server"
        # timeout.
        self.finished.wait(self.con._close_timeout)
        # Time to clean up.
        self.sock.shutdown(socket.SHUT_RDWR)
        self.sock.close()
        self.con._feeder = None


class TestFakeLogsContainer(unittest.TestCase):
    def mkcontainer(self, *args, **kw):
        kw.setdefault('close_timeout', 0.1)
        con = FakeLogsContainer(*args, **kw)
        self.addCleanup(con.cleanup)
        return con

    def stream(self, con, **kw):
        return list(stream_logs(con, **kw))

    def test_empty(self):
        """
        Streaming logs for a container with no logs returns immediately.
        """
        con = self.mkcontainer([])
        self.assertEqual(self.stream(con), [])
        self.assertEqual(con.logs(tail=1), b'')

    def test_tail_only_returns_streamed(self):
        """
        Only logs that have been streamed can be tailed.
        """
        con = self.mkcontainer([(0, b'hello\n'), (0, b'goodbye\n')])
        self.assertEqual(con.logs(tail=2), b'')
        self.assertEqual(self.stream(con), [b'hello\n', b'goodbye\n'])
        self.assertEqual(con.logs(tail=2), b'hello\ngoodbye\n')

    def test_tail_tails(self):
        """
        Tailing returns the last N log lines, or all line if there are fewer
        than N or if N is 'all' (which is the default).
        """
        con = self.mkcontainer([(0, b'hello\n'), (0, b'goodbye\n')])
        self.stream(con)
        self.assertEqual(con.logs(tail=1), b'goodbye\n')
        self.assertEqual(con.logs(tail=2), b'hello\ngoodbye\n')
        self.assertEqual(con.logs(tail=3), b'hello\ngoodbye\n')
        self.assertEqual(con.logs(tail='all'), b'hello\ngoodbye\n')
        self.assertEqual(con.logs(), b'hello\ngoodbye\n')

    def test_streaming_waits(self):
        """
        Streamed logs will be returned at specified intervals. Any logs that
        have already been streamed are not returned again if we set tail=0.

        NOTE: This test measures wall-clock time, so if something causes it to
              be too slow the second assertion may fail.
        """
        con = self.mkcontainer([(0.1, b'hello\n'), (0.2, b'goodbye\n')])
        t0 = datetime.now()
        self.assertEqual(self.stream(con), [b'hello\n', b'goodbye\n'])
        t1 = datetime.now()
        self.assertEqual(self.stream(con, tail=0), [])
        t2 = datetime.now()
        self.assertLess(0.3, (t1 - t0).total_seconds())
        self.assertLess((t2 - t1).total_seconds(), 0.3)


class TestStreamLogsFunc(unittest.TestCase):
    def mkcontainer(self, *args, **kw):
        kw.setdefault('close_timeout', 0.1)
        con = FakeLogsContainer(*args, **kw)
        self.addCleanup(con.cleanup)
        return con

    def stream_only(self, con, timeout=1):
        return list(stream_logs(con, tail=0, timeout=timeout))

    def stream_logs(self, con, timeout=0.5, **kw):
        return stream_logs(con, timeout=timeout, **kw)

    def test_stream_only(self):
        """
        If there are no historical logs, we get all the streamed logs.
        """
        con = self.mkcontainer([
            (0.1, b'hello\n'),
            (0.1, b'goodbye\n'),
        ])
        self.assertEqual(
            list(self.stream_logs(con, tail=0)), [b'hello\n', b'goodbye\n'])

    def test_historical_only(self):
        """
        If all the logs have been streamed, we only get the old ones.
        """
        con = self.mkcontainer([
            (0.1, b'hello\n'),
            (0.1, b'goodbye\n'),
        ])
        self.assertEqual(self.stream_only(con), [b'hello\n', b'goodbye\n'])
        self.assertEqual(self.stream_only(con), [])
        self.assertEqual(
            list(self.stream_logs(con)), [b'hello\n', b'goodbye\n'])

    def test_timeout_and_stream_again(self):
        """
        If we take too long to get the next line, we time out. When we stream
        again, we get historical logs as well as new ones.
        """
        con = self.mkcontainer([
            (0.1, b'hello\n'),
            (0.2, b'goodbye\n'),
        ])
        lines = []
        with self.assertRaises(TimeoutError):
            for line in self.stream_logs(con, timeout=0.15):
                lines.append(line)
        self.assertEqual(lines, [b'hello\n'])
        lines = list(self.stream_logs(con, timeout=0.35))
        self.assertEqual(lines, [b'hello\n', b'goodbye\n'])


class TestWaitForLogsMatchingFunc(unittest.TestCase):
    def mkcontainer(self, *args, **kw):
        kw.setdefault('close_timeout', 0.1)
        con = FakeLogsContainer(*args, **kw)
        self.addCleanup(con.cleanup)
        return con

    def wflm(self, con, matcher, timeout=0.5, **kw):
        return wait_for_logs_matching(con, matcher, timeout=timeout, **kw)

    def test_one_matching_line(self):
        """
        If one matching line is logged, all is happy.
        """
        con = self.mkcontainer([
            (0, b'hello\n'),
        ])
        # If this doesn't raise an exception, the test passes.
        self.wflm(con, EqualsMatcher('hello'))

    def test_one_non_matching_line(self):
        """
        If there's no match by the time the logs end, we raise an exception.
        """
        con = self.mkcontainer([
            (0, b'goodbye\n'),
        ])
        with self.assertRaises(RuntimeError) as cm:
            self.wflm(con, EqualsMatcher('hello'))
        self.assertIn(
            "Logs matching EqualsMatcher('hello') not found.",
            str(cm.exception))
        self.assertIn('goodbye\n', str(cm.exception))

    def test_timeout_first_line(self):
        """
        If we take too long to get the first line, we time out.
        """
        con = self.mkcontainer([
            (0.2, b'hello\n'),
        ])
        with self.assertRaises(TimeoutError) as cm:
            self.wflm(con, EqualsMatcher('hello'), timeout=0.1)
        self.assertIn(
            "Timeout (0.1s) waiting for logs matching EqualsMatcher('hello').",
            str(cm.exception))
        self.assertNotIn('hello\n', str(cm.exception))

    def test_timeout_later_line(self):
        """
        If we take too long to get a later line, we time out.
        """
        con = self.mkcontainer([
            (0, b'hi\n'),
            (0.2, b'hello\n'),
        ])
        with self.assertRaises(TimeoutError) as cm:
            self.wflm(con, EqualsMatcher('hello'), timeout=0.1)
        self.assertIn(
            "Timeout (0.1s) waiting for logs matching EqualsMatcher('hello').",
            str(cm.exception))
        self.assertIn('hi\n', str(cm.exception))
        self.assertNotIn('hello\n', str(cm.exception))

    def test_default_encoding(self):
        """
        By default, we assume logs are UTF-8.
        """
        con = self.mkcontainer([
            (0, b'\xc3\xbeorn\n'),
        ])
        # If this doesn't raise an exception, the test passes.
        self.wflm(con, EqualsMatcher('\u00feorn'))

    def test_encoding(self):
        """
        We can operate on logs that use excitingly horrible encodings.
        """
        con = self.mkcontainer([
            (0, b'\xfeorn\n'),
        ])
        # If this doesn't raise an exception, the test passes.
        self.wflm(con, EqualsMatcher('\u00feorn'), encoding='latin1')

    def test_kwargs(self):
        """
        We pass through any kwargs we don't recognise to docker.
        """
        con = self.mkcontainer([(0, b'hi\n')], {'stdout': False})
        with self.assertRaises(AssertionError):
            self.wflm(con, EqualsMatcher('hi'))
        with self.assertRaises(AssertionError):
            self.wflm(con, EqualsMatcher('hi'), stdout=False, stderr=False)
        self.wflm(con, EqualsMatcher('hi'), stdout=False)


class FakeAndRealContainerMixin:
    def wflm(self, con, matcher, timeout=0.5, **kw):
        return wait_for_logs_matching(con, matcher, timeout=timeout, **kw)

    def stream_logs(self, con, timeout=0.5, **kw):
        for line in stream_logs(con, timeout=timeout, **kw):
            yield line

    def test_fake_and_real_logging_behaviour(self):
        """
        Our fake logs container should exhibit similar behaviour to a real
        container.
        """
        logger = self.start_logging_container()
        self.wflm(logger, EqualsMatcher('Log entry 1'), timeout=1)

        early_logs = logger.logs().decode('utf8').splitlines()
        self.assertIn('Log entry 1', early_logs)
        self.assertNotIn('Log entry 4', early_logs)

        streamed_logs = []
        with self.assertRaises(TimeoutError):
            for line in self.stream_logs(logger, tail=0, timeout=0.7):
                streamed_logs.append(line.decode('utf8').rstrip())
        self.assertNotIn('Log entry 1', streamed_logs)
        self.assertIn('Log entry 4', streamed_logs)
        self.assertNotIn('Log entry 9', streamed_logs)

        streamed_logs = []
        for line in self.stream_logs(logger, timeout=2):
            streamed_logs.append(line.decode('utf8').rstrip())
        self.assertIn('Log entry 1', streamed_logs)
        self.assertIn('Log entry 4', streamed_logs)
        self.assertIn('Log entry 9', streamed_logs)


class TestWithFakeContainer(unittest.TestCase, FakeAndRealContainerMixin):
    def mkcontainer(self, *args, **kw):
        kw.setdefault('close_timeout', 0.1)
        con = FakeLogsContainer(*args, **kw)
        self.addCleanup(con.cleanup)
        return con

    def start_logging_container(self):
        return self.mkcontainer([
            (0.2, 'Log entry {}\n'.format(n).encode('utf8'))
            for n in range(10)])


@dockertest()
class TestWithRealContainer(unittest.TestCase, FakeAndRealContainerMixin):
    def setUp(self):
        self.dh = DockerHelper()
        self.addCleanup(self.dh.teardown)

    def start_logging_container(self):
        script = '\n'.join([
            'for n in 0 1 2 3 4 5 6 7 8 9; do',
            '    sleep 0.2',
            '    echo "Log entry ${n}"',
            'done',
        ])
        logger = self.dh.containers.create(
            'logger', IMG, command=['sh', '-c', script])
        self.addCleanup(self.dh.containers.remove, logger, force=True)
        logger.start()
        logger.reload()
        return logger



"""
This is a module for horrible low-level things that we really wish we
didn't have to build and maintain ourselves.
"""

import socket
import struct
import time
from contextlib import contextmanager

from urllib3.exceptions import ReadTimeoutError


# We want time.monotonic on Pythons that have it, otherwise time.time will have
# to do.
get_time = getattr(time, 'monotonic', time.time)


class SocketClosed(Exception):
    """
    Exception used for flow control. :-(
    """


def stream_logs(container, stdout=True, stderr=True, tail='all', timeout=10.0):
    """
    Stream logs from a Docker container within a timeout.

    We can't use docker-py's existing streaming support because that's stuck
    behind a blocking API and we have no (sane) way to enforce a timeout.
    """
    deadline = get_time() + timeout

    with _haxxed_mrsh(container):
        resp, sock = container.logs(
            stdout=stdout, stderr=stderr, stream=True, tail=tail, follow=True)
    try:
        while True:
            try:
                yield read_frame(resp, sock, deadline)
            except SocketClosed:
                return
    finally:
        # We also need to close the response object to avoid leaking any
        # resources.
        resp.close()


@contextmanager
def _haxxed_mrsh(container):
    """
    This temporarily monkey-patches the Docker API client's
    _multiplexed_response_stream_helper method so we can get the raw socket and
    do our own timeout-enabled streaming.
    """
    api = container.client.api

    def _mrsh(response):
        return response, api._get_raw_response_socket(response)._sock

    _orig_mrsh = api._multiplexed_response_stream_helper
    try:
        api._multiplexed_response_stream_helper = _mrsh
        yield container
    finally:
        api._multiplexed_response_stream_helper = _orig_mrsh


def read_by_deadline(resp, sock, deadline, n):
    """
    Read up to N bytes from a socket. If there's nothing to read, signal that
    it's closed.
    """
    time_left = deadline - get_time()
    # Avoid past-deadline special cases by setting a small timeout instead.
    sock.settimeout(max(time_left, 0.001))
    try:
        data = resp.raw.read(n)
    except (ReadTimeoutError, socket.timeout) as e:
        raise TimeoutError('Timeout waiting for container logs.')
    if len(data) == 0:
        raise SocketClosed()
    return data


def read_n_bytes(resp, sock, n, deadline):
    """
    Read exactly N bytes from a socket before a timeout deadline. We assume
    that the selector contains exactly one socket.
    """
    buf = b''
    while len(buf) < n:
        buf += read_by_deadline(resp, sock, deadline, n - len(buf))
    return buf


def read_frame(resp, sock, deadline):
    """
    Read a Docker stream frame before a timeout deadline.
    """
    header = read_n_bytes(resp, sock, 8, deadline)
    _, size = struct.unpack('>BxxxL', header)
    return read_n_bytes(resp, sock, size, deadline)



import docker

import pytest

from seaworthy.checks import docker_client
from seaworthy.definitions import ContainerDefinition
from seaworthy.helpers import DockerHelper, fetch_images
from seaworthy.pytest.checks import dockertest
from seaworthy.pytest.fixtures import (
    clean_container_fixtures, docker_helper_fixture, image_fetch_fixture,
    resource_fixture)


IMG = 'nginx:alpine'


def setup_module():
    with docker_client() as client:
        fetch_images(client, [IMG])


# We redefine the docker_helper fixture here to have function scope so our
# tests don't conflict.
docker_helper = docker_helper_fixture(scope='function')


@dockertest()
class TestDockerHelperFixture:
    def test_setup_teardown(self):
        """
        The fixture should yield a setup helper, and afterwards tear down the
        helper.
        """
        fixture_gen = docker_helper()
        helper = next(fixture_gen)
        assert isinstance(helper, DockerHelper)

        # Test we can create a container; if we can the helper must be set
        # up
        container = helper.containers.create('test', IMG)

        # Test things are torn down
        with pytest.raises(StopIteration):
            next(fixture_gen)

        with pytest.raises(docker.errors.NotFound):
            container.reload()


@dockertest()
class TestImageFetchFixtureFunc:
    def test_create_fixture(self, docker_helper):
        """
        We can create an fixture that fetches an image and returns an image
        model.
        """
        fixture = image_fetch_fixture(IMG, name='image', scope='module')
        image = fixture(docker_helper)

        assert isinstance(image, docker.models.images.Image)
        assert IMG in image.tags


@dockertest()
class TestContainerFixtureFunc:
    def test_setup_teardown(self, docker_helper):
        """
        The fixture should yield a started container, and afterwards stop and
        remove the container.
        """
        fixture = resource_fixture(
            ContainerDefinition(name='test', image=IMG), 'test')
        fixture_gen = fixture(docker_helper)
        container = next(fixture_gen)

        assert isinstance(container, ContainerDefinition)
        assert container.inner().status == 'running'

        # Test things are torn down
        with pytest.raises(StopIteration):
            next(fixture_gen)

        # Container has been stopped and removed
        with pytest.raises(RuntimeError):
            container.inner()


@dockertest()
class TestCleanContainerFixturesFunc:
    def test_setup_teardown(self, docker_helper):
        """
        The fixture should yield a started container, and afterwards stop and
        remove the container.
        """
        raw_fixture, fixture = clean_container_fixtures(
            ContainerDefinition(name='test', image=IMG), 'test')
        fixture_gen = raw_fixture(docker_helper)
        # TODO: Assert on cleaning fixture
        container = next(fixture_gen)

        assert isinstance(container, ContainerDefinition)
        assert container.inner().status == 'running'

        # Test things are torn down
        with pytest.raises(StopIteration):
            next(fixture_gen)

        # Container has been stopped and removed
        with pytest.raises(RuntimeError):
            container.inner()
